#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <algorithm>

#include "plugin.hpp"

#include "types.hpp"
#include "types_xhrip.hpp"
#include "manager.hpp"
#include "detector.hpp"
#include "tools.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Detector*         detector;
extern Options           op;

HANDLE Manager::open(int fNum)
{
  if(fNum >= noVolumes) return INVALID_HANDLE_VALUE;
  
  return CreateFileW(volumeName[fNum],
                    GENERIC_READ,
                    FILE_SHARE_READ | FILE_SHARE_WRITE,
                    NULL,
                    OPEN_EXISTING,
                    FILE_FLAG_SEQUENTIAL_SCAN,
                    NULL);
}

bool Manager::setPointer(DWORD offset)
{
  int vNum = 0;
  for(; vNum < noVolumes; ++vNum)
  {
    if(volumeSize[vNum] < offset)
      offset -= volumeSize[vNum];
    else
      break;
  }
  if(vNum == noVolumes) return false;
  if(vNum == volumeNum)
  {
    SetFilePointer(hostFile, offset + volumeOffset[vNum], NULL, FILE_BEGIN);
    return true;
  }
  CloseHandle(hostFile);

  hostFile = open(vNum);
  if(hostFile == INVALID_HANDLE_VALUE) return false;
  SetFilePointer(hostFile, offset + volumeOffset[vNum], NULL, FILE_BEGIN);
  volumeNum = vNum;
  return true;
}

bool Manager::read(void* buf, WORD size)
{
  DWORD noBytesRead;
  ReadFile(hostFile, buf, size, &noBytesRead, NULL);
  if(noBytesRead == size) return true;
  
  if(++volumeNum == noVolumes) return false;
  
  CloseHandle(hostFile);
  hostFile = open(volumeNum);
  
  if(hostFile == INVALID_HANDLE_VALUE) return false;
  SetFilePointer(hostFile, volumeOffset[volumeNum], NULL, FILE_BEGIN);
  ReadFile(hostFile, (BYTE*)buf + noBytesRead, size - noBytesRead, &noBytesRead, NULL);

  return true;
}

Manager::Manager(const wchar_t* fileName, int offset)
{
  lstrcpyW(hostFileName, fileName);
  hdrOffset = offset;
  noFiles   = 0;
  volumeNum = 0;
  
  ZeroMemory(&lastModifed, sizeof(WIN32_FIND_DATA));

  HANDLE hFile = CreateFileW(hostFileName,
                            GENERIC_READ,
                            FILE_SHARE_READ | FILE_SHARE_WRITE,
                            NULL,
                            OPEN_EXISTING,
                            FILE_FLAG_SEQUENTIAL_SCAN,
                            NULL);
  DWORD fileSize = GetFileSize(hFile, NULL);
  CloseHandle(hFile);

  volumeSize  [0] = fileSize - hdrOffset;
  volumeName[0] = new wchar_t[300]{};
  volumeOffset[0] = hdrOffset;
  lstrcpyW(volumeName[0], hostFileName);

  wchar_t name[300];
  lstrcpyW(name, hostFileName);
  wchar_t *ptr = name + lstrlenW(name);
  while(ptr != name && *ptr != '.') --ptr;
  if(ptr!=name) *(ptr+2) = 0;
  lstrcatW(name, L"*");
  
  for(noVolumes = 1; noVolumes < 16; noVolumes++)
  {
    bool found = false;

    WIN32_FIND_DATAW data;
    HANDLE h = FindFirstFileW(name, &data);
    if(h == INVALID_HANDLE_VALUE) break;

    do
    {
      wchar_t tmpName[300];
      lstrcpyW(tmpName, hostFileName);
      lstrcpyW(pointToName(tmpName), data.cFileName);

      HANDLE volume = CreateFileW(tmpName,
                                 GENERIC_READ,
                                 FILE_SHARE_READ | FILE_SHARE_WRITE,
                                 NULL,
                                 OPEN_EXISTING,
                                 FILE_FLAG_SEQUENTIAL_SCAN,
                                 NULL);
      if(volume == INVALID_HANDLE_VALUE) break;

      HoHdr hdr;
      DWORD noBytesRead;

      ReadFile(volume, &hdr, sizeof(HoHdr), &noBytesRead, NULL);
      CloseHandle(volume);

      if(hdr.checkSum != calculateCheckSum(hdr)) continue;

      char type = '0' + noVolumes - 1;
      if(type > '9') type += 'A' - '9' - 1;
      if(hdr.type == type && !memcmp(&hdr.start, "rp", 2))
      {
        found = true;
        volumeSize  [noVolumes] = data.nFileSizeLow - sizeof(HoHdr);
        volumeOffset[noVolumes] = sizeof(HoHdr);
        volumeName[noVolumes] = new wchar_t[300]{};
        lstrcpyW(volumeName[noVolumes], tmpName);
        break;
      }
    }while(FindNextFileW(h, &data));

    FindClose(h);
    if(!found) break;
  }
}

Manager::~Manager()
{
  for(int i = 0; i < noVolumes; ++i)
    delete[] volumeName[i];
}

void Manager::GetOpenPanelInfo(struct OpenPanelInfo *info)
{
  info->StructSize    = sizeof(*info);
  info->Flags         = OPIF_ADDDOTS | OPIF_SHOWNAMESONLY | OPIF_SHOWPRESERVECASE;
  info->HostFile      = hostFileName;
  info->CurDir        = L"";
  info->Format        = L"HRiP";
  info->StartSortMode = SM_UNSORTED;
  
  static wchar_t panelTitle[260];
  panelTitle[0] = 0;
  lstrcatW(panelTitle, L" HRiP:");
  lstrcatW(panelTitle, pointToName(hostFileName));
  lstrcatW(panelTitle, L" ");
  info->PanelTitle = panelTitle;

  static PanelMode   mode[10];
  static const wchar_t*       columnTitles3[4];
  columnTitles3[0] = GetMsg(MName);
  columnTitles3[1] = GetMsg(MSectorSize);
  columnTitles3[2] = GetMsg(MName);
  columnTitles3[3] = GetMsg(MSectorSize);

  static const wchar_t*       columnTitles4[5];
  columnTitles4[0] = GetMsg(MName);
  columnTitles4[1] = GetMsg(MSize);
  columnTitles4[2] = GetMsg(MStart);
  columnTitles4[3] = GetMsg(MSectorSize);
  columnTitles4[4] = GetMsg(MPackedSize);
  
  static const wchar_t*       columnTitles5[2];
  columnTitles5[0] = GetMsg(MName);
  columnTitles5[1] = GetMsg(MFormat);

  static const wchar_t*       columnTitles6[2];
  columnTitles6[0] = GetMsg(MName);
  columnTitles6[1] = GetMsg(MComment);
  
  // C0 - trdos name
  // C1 - trdos 'start'
  // C2 -   
  // C3 - packedSize
  // C4 -  
  // C5 -   
  
  mode[3].ColumnTypes=L"N,C2,N,C2";
  mode[3].ColumnWidths=L"0,3,0,3";
  mode[3].ColumnTitles=columnTitles3;
  mode[3].StatusColumnTypes=L"C0,S,C2";
  mode[3].StatusColumnWidths=L"0,5,3";
  mode[3].Flags = PMFLAGS_ALIGNEXTENSIONS;
  
  mode[4].ColumnTypes=L"C0,S,C1,C2,C3";
  mode[4].ColumnWidths=L"0,5,5,3,5";
  mode[4].StatusColumnTypes=L"N,S,C2";
  mode[4].StatusColumnWidths=L"0,5,3";
  mode[4].ColumnTitles=columnTitles4;

  mode[5].ColumnTypes=L"C0,C4";
  mode[5].ColumnWidths=L"12,0";
  mode[5].StatusColumnTypes=L"N,S,C2";
  mode[5].StatusColumnWidths=L"0,5,3";
  mode[5].ColumnTitles=columnTitles5;

  mode[6].ColumnTypes=L"C0,C5";
  mode[6].ColumnWidths=L"12,0";
  mode[6].StatusColumnTypes=L"N,S,C2";
  mode[6].StatusColumnWidths=L"0,5,3";
  mode[6].ColumnTitles=columnTitles6;
  
  info->PanelModesArray = mode;
  info->PanelModesNumber = sizeof(mode)/sizeof(mode[0]);
  info->StartPanelMode = '0' + op.defaultPanelMode;
}

int Manager::getFindData(PluginPanelItem **pPanelItem, size_t *pNoItems, int opMode)
{
  if(!readInfo()) return FALSE;

  PluginPanelItem *item = new PluginPanelItem[noFiles]{};
  *pNoItems   = noFiles;
  *pPanelItem = item;

  if(noFiles == 0) return TRUE;

  //      trdos' 
  PanelInfo panel{ sizeof(PanelInfo) };
  startupInfo.PanelControl(this, FCTL_GETPANELINFO, 0, &panel);
  int nameColumnWidth = 12;

  if(panel.ViewMode == 4)
  {
    // 21 -   ,  2 -  
    nameColumnWidth = panel.PanelRect.right - panel.PanelRect.left - 2 - 21;
  }
  if(nameColumnWidth < 12) nameColumnWidth = 12;
  
  wchar_t* name = new wchar_t[nameColumnWidth+1];

  for(int fNum = 0; fNum < noFiles; ++fNum)
  {
    item[fNum].FileName = pcFiles[fNum].name;
    item[fNum].FileSize = files[fNum].size;
    
    item[fNum].CustomColumnNumber = 6;
    wchar_t **CustomColumnData = new LPWSTR[item[fNum].CustomColumnNumber]{};
    item[fNum].CustomColumnData = CustomColumnData;

    CustomColumnData[0] = new wchar_t[nameColumnWidth + 1]{};

    std::fill_n(name, nameColumnWidth+1, L' ');

    makeTrDosName(name, files[fNum], op.showExt, nameColumnWidth);
    lstrcpyW(CustomColumnData[0], name);

    CustomColumnData[1] = new wchar_t[6];
    wsprintfW(CustomColumnData[1], L"%5d", files[fNum].start);

    CustomColumnData[2] = new wchar_t[4];
    wsprintfW(CustomColumnData[2],
             L"%3d",
             files[fNum].noSecs);

    CustomColumnData[3] = new wchar_t[6];
    wsprintfW(CustomColumnData[3], L"%5d", infos[fNum].packedSize);

    const wchar_t* description = detector->description(pcFiles[fNum].type);
    if(description)
    {
      CustomColumnData[4] = new wchar_t[lstrlenW(description)+1];
      lstrcpyW(CustomColumnData[4], description);
    }
    else
    {
      CustomColumnData[4] = new wchar_t[1];
      CustomColumnData[4][0] = 0;
    }
    
    const wchar_t* comment = pcFiles[fNum].comment;
    if(comment[0])
    {
      CustomColumnData[5] = new wchar_t[lstrlenW(comment)+1];
      lstrcpyW(CustomColumnData[5], comment);
    }
    else
    {
      CustomColumnData[5] = new wchar_t[1];
      CustomColumnData[5][0] = 0;
    }
  }
  delete[] name;
  return TRUE;
}

void Manager::freeFindData(PluginPanelItem *panelItem, int noItems)
{
  for (int i = 0; i < noItems; i++)
  {
    delete[] panelItem[i].CustomColumnData[0];
    delete[] panelItem[i].CustomColumnData[1];
    delete[] panelItem[i].CustomColumnData[2];
    delete[] panelItem[i].CustomColumnData[3];
    delete[] panelItem[i].CustomColumnData[4];
    delete[] panelItem[i].CustomColumnData[5];
    delete[] panelItem[i].CustomColumnData;
  }
  delete[] panelItem;
}

bool Manager::readInfo(void)
{
  //       
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(hostFileName, &data);
  if(h == INVALID_HANDLE_VALUE) return false;
  FindClose(h);
  
  if(CompareFileTime(&data.ftLastWriteTime, &lastModifed.ftLastWriteTime) != 0 ||
     data.nFileSizeLow != lastModifed.nFileSizeLow)
  {
    lastModifed = data;
    if(!openHostFile()) return false;

    SetFilePointer(hostFile, hdrOffset+3, NULL, FILE_BEGIN);
    DWORD noBytesRead;
    ReadFile(hostFile, &noFiles, 1, &noBytesRead, NULL);
    if(noFiles == 'P') noFiles = 0xFF; //  HRiP  0.98

    BlockHdr block;
    ExtInfo  info;
    //      

    DWORD offset = 8;
    setPointer(offset);
    bool isUnexpectedEnd = false;
    for(int i = 0; i < noFiles && !isUnexpectedEnd; ++i)
    {
      WORD pSize = 0;
      infos[i].offset = offset;
      do
      {
        ZeroMemory(&block, sizeof(BlockHdr));
        read(&block, 6);
        
        char signature[] = { 'H', 'r', 's', 't', '2' };
        if(memcmp(&block.signature, signature, sizeof(signature)))
        {
          if(noFiles != 0xFF) //  HRiP  0.98
          {
            errorMessage(GetMsg(MEndOfArchive));
            isUnexpectedEnd = true;
          }
          noFiles = i;
          break;
        }
        
        //  
        if(block.flags & 0x80)
        {
          BYTE dirHdrSize, dirHdr[255];
          read(&dirHdrSize, 1);
          read(dirHdr, dirHdrSize);
          offset += 7 + dirHdrSize;
          setPointer(offset);
          infos[i].offset = offset;
          continue;
        }
        
        //    
        read(&block.size, 5);
        
        pSize += block.pSize;

        if(block.extInfoSize >= 18)
        {
          read(&info, sizeof(ExtInfo));
          CopyMemory(&files[i], &info.file, sizeof(SCLHdr));
        }
        //     ,
        //  . .
        // !!!!       - 
        
        offset += sizeof(BlockHdr) + block.extInfoSize + block.pSize;
        setPointer(offset);
      } while(!(block.flags & 0x02));
      
      infos[i].packedSize = pSize;
    }
    makePCNames();
    closeHostFile();
  }
  return true;
}
